﻿<#
    .SYNOPSIS
     Sample Powershell script to deploy a VMware Access Point virtual appliance using ovftool.
    .EXAMPLE
     .\apdeployhyperv.ps1 -iniFile ap1.ini 
#>

param([string]$iniFile = "ap.ini")

#
# Load the dependent PowerShell Module
#

$ScriptPath = $MyInvocation.MyCommand.Path
$ScriptDir  = Split-Path -Parent $ScriptPath
$apDeployModule=$ScriptDir+"\uagdeploy.psm1"

if (!(Test-path $apDeployModule)) {
	WriteErrorString "Error: PowerShell Module $apDeployModule not found."
	Exit
}

import-module $apDeployModule -Force

function AddKVPUnit {
    param($VMName, $key, $value)
    
    #
    # Add Key-Value Pairs for the VM
    #

    #if ($key.Contains("Password")) {
    #    Write-Host "Setting $key=******"
    #} else {
    #    Write-Host "Setting $key=$value"
    #}

    $VmMgmt = gwmi -n "Root\Virtualization\V2" Msvm_VirtualSystemManagementService #Get-WmiObject -Namespace root\virtualization\v2 -Class Msvm_VirtualSystemManagementService
    $Vm = gwmi -n "root\virtualization\v2" Msvm_ComputerSystem|?{$_.ElementName -eq $VMName }  #Get-WmiObject -Namespace root\virtualization\v2 -Class Msvm_ComputerSystem -Filter {ElementName='TEST-APP38'}     # has to be same as $VMName

    $kvpDataItem = ([WMIClass][String]::Format("\\{0}\{1}:{2}", $VmMgmt.ClassPath.Server, $VmMgmt.ClassPath.NamespacePath, "Msvm_KvpExchangeDataItem")).CreateInstance()
    $null=$KvpItem.psobject.properties
    $kvpDataItem.Name = $key
    $kvpDataItem.Data = $value
    $kvpDataItem.Source = 0
    $result = $VmMgmt.AddKvpItems($Vm, $kvpDataItem.PSBase.GetText(1))

    $job = [wmi]$result.Job

    if (!$job) {
        WriteErrorString "Error: Failed to set KVP $key on $VMName"
        Return
    }    

    if ($job) {
        $job.get()
        #write-host $job.jobstate
        #write-host $job.SystemProperties.Count.ToString()
    }

    while($job.jobstate -lt 7) {
	    $job.get()
        Start-Sleep -Seconds 2
    } 

    if ($job.ErrorCode -ne 0) {
        WriteErrorString "Error: Failed to set KVP $key on $VMName (error code $($job.ErrorCode))"
        Return
    }    
    
    if ($job.Status -ne "OK") {
        WriteErrorString "Error: Failed to set KVP $key on $VMName (status $job.Status)"
        Return
    }

    $job
}

function GetKVP {
    param($VMName, $key)

    $VmMgmt = gwmi -n "Root\Virtualization\V2" Msvm_VirtualSystemManagementService #Get-WmiObject -Namespace root\virtualization\v2 -Class Msvm_VirtualSystemManagementService
    $Vm = gwmi -n "root\virtualization\v2" Msvm_ComputerSystem|?{$_.ElementName -eq $VMName }  #Get-WmiObject -Namespace root\virtualization\v2 -Class Msvm_ComputerSystem -Filter {ElementName='TEST-APP38'}     # has to be same as $VMName

    $n = $vm.GetRelated("Msvm_KvpExchangeComponent").GuestIntrinsicExchangeItems
    $n = $vm.GetRelated("Msvm_KvpExchangeComponent").GuestExchangeItems

    $n = $vm.GetRelated("Msvm_KvpExchangeComponent").GetRelated('Msvm_KvpExchangeComponentSettingData').HostExchangeItems

    $n | % {
        $GuestExchangeItemXml = ([XML]$_).SelectSingleNode(`
            "/INSTANCE/PROPERTY[@NAME='Name']/VALUE[child::text()='$key']")

        if ($GuestExchangeItemXml -ne $null)
        {
            $val = $GuestExchangeItemXml.SelectSingleNode( `
                "/INSTANCE/PROPERTY[@NAME='Data']/VALUE/child::text()").Value
                $val
                Return
        }
    }
}

function AddKVP {
    param($VMName, $key, $value)
    $max = 1000
    $len=$value.length
    $index=0
    if ($len -le $max) {
        $job=AddKVPUnit $VMName $key $value
    } else {
        for ($i=0; $i -lt $len; $i += $max) {
            $chunkSize = [Math]::Min($max, ($len - $i))
            $valueChunk=$value.Substring($i, $chunkSize)
            $keyChunk=$key+"."+$index.ToString(0)
            $job=AddKVPUnit $VMName $keyChunk $valueChunk
            $index++
        }
    }
    $job
}

function DeleteKVP {
    param($VMName, $key)

    $VmMgmt = gwmi -n "Root\Virtualization\V2" Msvm_VirtualSystemManagementService
    $Vm = gwmi -n "root\virtualization\v2" Msvm_ComputerSystem|?{$_.ElementName -eq $VMName }

    $kvpDataItem = ([WMIClass][String]::Format("\\{0}\{1}:{2}", $VmMgmt.ClassPath.Server, $VmMgmt.ClassPath.NamespacePath, "Msvm_KvpExchangeDataItem")).CreateInstance()
    $null=$KvpItem.psobject.properties
    $kvpDataItem.Name = $key
    $kvpDataItem.Data = [String]::Empty
    $kvpDataItem.Source = 0
    $result = $VmMgmt.RemoveKvpItems($Vm, $kvpDataItem.PSBase.GetText(1))

    $job = [wmi]$result.Job

    if (!$job) {
        WriteErrorString "Error: Failed to set KVP $key on $VMName"
        Return
    }    

    if ($job) {
        $job.get()
        #write-host $job.jobstate
        #write-host $job.SystemProperties.Count.ToString()
    }

    while($job.jobstate -lt 7) {
	    $job.get()
        Start-Sleep -Seconds 2
    } 

    if ($job.ErrorCode -ne 0) {
        WriteErrorString "Error: Failed to set KVP $key on $VMName (error code $($job.ErrorCode))"
        Return
    }    
    
    if ($job.Status -ne "OK") {
        WriteErrorString "Error: Failed to set KVP $key on $VMName (status $job.Status)"
        Return
    }

    $job
}

function DeleteKVPAll {
    param($VMName)

    $VmMgmt = gwmi -n "Root\Virtualization\V2" Msvm_VirtualSystemManagementService
    $Vm = gwmi -n "root\virtualization\v2" Msvm_ComputerSystem|?{$_.ElementName -eq $VMName }

    $hostExchangeItems = $vm.GetRelated("Msvm_KvpExchangeComponent").GetRelated('Msvm_KvpExchangeComponentSettingData').HostExchangeItems

    $hostExchangeItems | % {

    $GuestExchangeItemXml = ([XML]$_).SelectSingleNode(`
        "/INSTANCE/PROPERTY[@NAME='Name']/VALUE")
        $key = $GuestExchangeItemXml.InnerText

        if ($key.length -gt 0) {
            $job = DeleteKVP $VMName $key
        }      
    }
}

function IsVMDeployed {
    param ($VMName, $ipAddress)
    #
    # WE consider the VM to be deployed if we can obtain an IP address from it and the ip address matches what is configured.
    #

    $out=Get-VM $VMName | ?{$_.ReplicationMode -ne “Replica”} | Select -ExpandProperty NetworkAdapters | Select IPAddresses
    if ($ipAddress.length -gt 0) {
        #static IP address
        if (($out.IPAddresses)[0] -eq $ipAddress) {
            return $true
        }
    } else {
        #DHCP
        if  ((($out.IPAddresses)[0]).length -gt 0) {
            $wait = 0
            #sleep for 4 mins (240 secs) to give time for the appliance to be ready
            while ($wait -le 240) {
                Write-Host -NoNewline "."
                $wait += 2
                Start-Sleep -Seconds 2
            }
            return $true
        }
    }

    return $false
}

function IsVMUp {
    param ($VMName, [ref]$ipAddress)

    #
    # WE consider the VM to be up if we can obtain an IP address from it.
    #

    $out=Get-VM $VMName | ?{$_.ReplicationMode -ne “Replica”} | Select -ExpandProperty NetworkAdapters | Select IPAddresses
    if ($out.IPAddresses.Length -gt 0) {
        $ipAddress = ($out.IPAddresses)[0]
        return $true
    }

    return $false
}


Write-host "Unified Access Gateway (UAG) virtual appliance deployment script"

$settings = ImportIni $iniFile

if (!(Test-path $iniFile)) {
	WriteErrorString "Error: Configuration file ($iniFile) not found."
	Exit
}

$VMName=$settings.General.name

#
# Assign and validate network settings
#

if ($settings.General.netInternet.length -eq 0) {
	WriteErrorString "Error: Missing netInternet Hyper-V switch name in .ini file"
    Exit
}

$dns=$settings.General.dns
$defaultGateway=$settings.General.defaultGateway

$netInternet=$settings.General.netInternet
$ip0=$settings.General.ip0

$netManagementNetwork=$settings.General.netManagementNetwork
$ip1=$settings.General.ip1

$netBackendNetwork=$settings.General.netBackendNetwork
$ip2=$settings.General.ip2

$gateway0=$settings.General.gateway0
$gateway1=$settings.General.gateway1
$gateway2=$settings.General.gateway2

$netmask0=$settings.General.netmask0
$netmask1=$settings.General.netmask1
$netmask2=$settings.General.netmask2

$routes0=$settings.General.routes0
$routes1=$settings.General.routes1
$routes2=$settings.General.routes2

if ((!$ip0) -And (!$ip1) -And (!$ip2)) {

	#
	# No IP addresses specified so we will use DHCP for address allocation
	#

	$ipAllocationPolicy = "dhcpPolicy"

} else {

	$ipAllocationPolicy = "fixedPolicy"

}

$deploymentOption=$settings.General.deploymentOption

if (!$deploymentOption) {
	$deploymentOption="onenic"
}

if ($ipAllocationPolicy -eq "fixedPolicy") {
	switch ($deploymentOption) {

		onenic {
			if (!$ip0) {
				WriteErrorString "Error: ip0 in the [General] section of $iniFile is missing."
				Exit
			}

		}
		twonic {
			if (!$ip0) {
				WriteErrorString "Error: ip0 in the [General] section of $iniFile is missing."
				Exit
			}
			if (!$ip1) {
				WriteErrorString "Error: ip1 in the [General] section of $iniFile is missing."
				Exit
			}
            # check for gateway and netmask as well...
		}
		threenic {
			if (!$ip0) {
				WriteErrorString "Error: ip0 in the [General] section of $iniFile is missing."
				Exit
			}
			if (!$ip1) {
				WriteErrorString "Error: ip1 in the [General] section of $iniFile is missing."
				Exit
			}
			if (!$ip2) {
				WriteErrorString "Error: ip2 in the [General] section of $iniFile is missing."
				Exit
			}
		}
		default {
			WriteErrorString "Error: deploymentOption=$deploymentOption in the [General] section of $iniFile is not valid."
			WriteErrorString "It must be set to onenic, twonic or threenic."
			Exit

		}
	}
}

if ($ip0) {
    if (!$gateway0) {
        WriteErrorString "Error: gateway0 in the [General] section of $iniFile is missing."
        Exit
    }

    if (!$netmask0) {
        WriteErrorString "Error: netmask0 in the [General] section of $iniFile is missing."
        Exit
    }
}

if ($ip1) {
    if (!$gateway1) {
        WriteErrorString "Error: gateway1 in the [General] section of $iniFile is missing."
        Exit
    }

    if (!$netmask1) {
        WriteErrorString "Error: netmask1 in the [General] section of $iniFile is missing."
        Exit
    }
}

if ($ip2) {
    if (!$gateway2) {
        WriteErrorString "Error: gateway2 in the [General] section of $iniFile is missing."
        Exit
    }

    if (!$netmask2) {
        WriteErrorString "Error: netmask2 in the [General] section of $iniFile is missing."
        Exit
    }
}

#
# Create New Virtual Machine
#

if ($VMName.length -gt 32) { 
	WriteErrorString "Error: Virtual machine name must be no more than 32 characters in length"
	Exit
}

if (!$VMName) {
	$VMName = VMName
}

$rootPwd = GetRootPwd $VMName
$adminPwd = GetAdminPwd $VMName

$settingsJSON=GetJSONSettings $settings

$vCPU=2
$vRAM=2GB
$VHDSize=10GB

$progresspreference = "SilentlyContinue"

#
# Stop existing VM of the same name
#

$out=Stop-VM -Name $VMName -TurnOff -Force -ErrorAction SilentlyContinue -WarningAction SilentlyContinue -ErrorVariable error -WarningVariable warning
if ((!$error) -And (!$warning)) {
    Write-Host "Powered off VM: $VMName"
}

#
# Delete existing VM of the same name
#

$out=Remove-VM -Name $VMName -Force -ErrorAction SilentlyContinue -WarningAction SilentlyContinue -ErrorVariable error -WarningVariable warning
if ((!$error) -And (!$warning)) {
    Write-Host "Deleted VM: $VMName"
}

$vhdxPath=$settings.General.ds

$out=mkdir $vhdxPath -ErrorAction SilentlyContinue

if (!(Test-path $vhdxPath)) {
	WriteErrorString "Error: unable to create data store folder $vhdxPath - check the datastore (ds) value in $iniFile"
	Exit
}

#
# Copy the initial vhdx image file
#

Write-Host "Copying $VMName starter disk to $vhdxPath"

$vhdxSource=$settings.General.source
$vhdxFile=$vhdxPath
$vhdxFile+="\"
$vhdxFile+=$VMName
$vhdxFile+=".vhdx"

cmd /c copy /z $settings.General.source $vhdxFile

if (!(Test-path $vhdxFile)) {
	WriteErrorString "Error: Failed to copy to $vhdxSource to $vhdxFile"
	Exit
}

#
# Create the new VM
#

Write-Host -NoNewline "Creating new VM: $VMName .."

$out=New-VM -Name $VMName -MemoryStartupBytes $vRAM -VHDPath $vhdxFile -SwitchName $netInternet -ErrorVariable Error -ErrorAction SilentlyContinue
if ($error) {
    Write-Host ". FAILED"
	WriteErrorString "Error: Failed create Access Point $VMName ($Error)"
    Exit
}

Write-Host ". OK"

#
# Add additional NICs if needed
#

if ($netManagementNetwork) {
    Add-VMNetworkAdapter -VMName $VMName -SwitchName $netManagementNetwork -Name "Management Network Adapter" -ErrorVariable Error -ErrorAction SilentlyContinue

    if ($error) {
	    WriteErrorString "Error: Failed to set Management Network NIC $netManagementNetwork - ($Error)"
        Exit
    }
}

if ($netBackendNetwork) {
    Add-VMNetworkAdapter -VMName $VMName -SwitchName $netBackendNetwork -Name "Backend Network Adapter" -ErrorVariable Error -ErrorAction SilentlyContinue
    if ($error) {
	    WriteErrorString "Error: Failed to set Backend Network NIC $netBackendNetwork - ($Error)"
        Exit
    }
}

#
# Start the new VM
#

$out=start-vm -vmname $VMName

# 
# Wait for KVP Exchange Service to start
#

Write-Host -NoNewline "Starting Guest Data Exchange Service ."

$KVPExchangeServiceStatus=Get-VMIntegrationService -VMName $VMName | Where-Object {$_.Name -eq 'Key-Value Pair Exchange'} | Where-Object {$_.PrimaryStatusDescription -eq 'OK'}
 while (! $KVPExchangeServiceStatus) {
    Write-Host -NoNewline "."
    Start-Sleep -Seconds 5 
    $KVPExchangeServiceStatus=Get-VMIntegrationService -VMName $VMName | Where-Object {$_.Name -eq 'Key-Value Pair Exchange'} | Where-Object {$_.PrimaryStatusDescription -eq 'OK'}
 }

Write-Host ". OK"

#
# Poll the VM to wait for it to be up then wait 20 more seconds
#

Write-Host -NoNewline "Waiting for Guest Integration Services to start on $VMName ."

$ipAddress=""
while ($true) {
    Write-Host -NoNewline "."
    if (IsVMUp ($VMName, [ref]$ipAddress)) {
        Start-Sleep 20
        Break
    }
    Start-Sleep -Seconds 5 
}

Write-Host ". OK"

#
# Add Key-Value Pairs for Virtual Machine
#

Write-Host -NoNewline "Setting initial $VMName configuration settings "

$job=AddKVP $VMName "rootPassword" $rootPwd
Write-Host -NoNewline "."

if ($adminPwd.length -gt 0) {
    $job=AddKVP $VMName "adminPassword" $adminPwd
    Write-Host -NoNewline "."
}

if ($dns.length -gt 0) {
    $job=AddKVP $VMName "DNS" $dns
    Write-Host -NoNewline "."
}

if ($defaultGateway.length -gt 0) {
    $job=AddKVP $VMName "defaultGateway" $defaultGateway
    Write-Host -NoNewline "."
}

if ($ip0) {
    $job=AddKVP $VMName "ip0" $ip0
    Write-Host -NoNewline "."
    $job=AddKVP $VMName "gateway0" $gateway0
    Write-Host -NoNewline "."
    $job=AddKVP $VMName "netmask0" $netmask0
    Write-Host -NoNewline "."
}

if ($ip1) {
    $job=AddKVP $VMName "ip1" $ip1
    Write-Host -NoNewline "."
    $job=AddKVP $VMName "gateway1" $gateway1
    Write-Host -NoNewline "."
    $job=AddKVP $VMName "netmask1" $netmask1
    Write-Host -NoNewline "."
}

if ($ip2) {
    $job=AddKVP $VMName "ip2" $ip2
    Write-Host -NoNewline "."
    $job=AddKVP $VMName "gateway2" $gateway2
    Write-Host -NoNewline "."
    $job=AddKVP $VMName "netmask2" $netmask2
    Write-Host -NoNewline "."
}

if ($routes0) {
    $job=AddKVP $VMName "routes0" $routes0
}

if ($routes1) {
    $job=AddKVP $VMName "routes1" $routes1
}

if ($routes2) {
    $job=AddKVP $VMName "routes2" $routes2
}

if ($settings.General.syslogUrl.length -gt 0) {
    $job=AddKVP $VMName "syslogUrl" $settings.General.syslogUrl
    Write-Host -NoNewline "."
}

$job=AddKVP $VMName "settingsJSON" $settingsJSON

Write-Host ". OK"

if($ipAddress -eq $ip0) {
    # Race Condition !!!
    # IPAddress that the AP got from DHCP is the same as has been configured.
    # Sleep for 3 minutes to give enough time for vmware_ap_sysconfig to be executed
    wait = 0
    while ($wait -le 180) {
        Write-Host -NoNewline "."
        $wait++
        Start-Sleep -Seconds 1
    }
} else {
    $wait = 0
    Write-Host -NoNewline "Applying $VMName configuration settings "
    while(-Not (IsVMDeployed $VMName $ip0 )) {
    Write-Host -NoNewline "."
        if($wait -ge 600) {
            # If its been more than 10 minutes, and VM has still not come up then, deployment failure
            write-Host ". Failed"
            WriteErrorString "UAG virtual appliance $VMName failed to deploy"
            DeleteKVPAll $VMName
            exit
        }
        # Wait utill deployment successful
        $wait++
        Start-Sleep -Seconds 1
    }
}
Write-Host ". OK"

Write-Host -NoNewline "Completing $VMName deployment .."
# Give it 10 seconds for vmware_ap_sysconfig to run
Start-Sleep -Seconds 10
Write-Host ". OK"
DeleteKVPAll $VMName
Write-host "UAG virtual appliance $VMName deployed successfully"
